#include <stdio.h>
#include <string.h>
#include <math.h>
#include <gtk/gtk.h>
#include <gdk_imlib.h>

#define GRAB_NONE 0
#define GRAB_XYO  1
#define GRAB_RO   2
#define GRAB_XYI  3
#define GRAB_RI   4

#define CURSW     5

int xi,yi,ri,xo,yo,ro,dt=500,dr=150;

GtkWidget *widget_src, *widget_dest;
GdkPixmap *pix_src=NULL;
GdkImage *im_src=NULL;
GdkImage *im_dest=NULL;
int src_grab=GRAB_NONE;



/* Src & dest image management */


guint32 rgb_to_pixel(guint32 rgb)
{
	static GdkVisual *vis=NULL;
	guint32 pixel=0;
	
	if (!vis) vis=gdk_visual_get_system();
	
	switch (vis->green_prec)
	{
		/*              case 0: paletted */
		case 5: /* RGB = 555 */
			pixel =(rgb & 0xF80000) >> 9;
			pixel+=(rgb & 0x00F800) >> 6;
			pixel+=(rgb & 0x0000F8) >> 3;
			break;
		case 6: /* RGB = 565 */
			pixel =(rgb & 0xF80000) >> 8;
			pixel+=(rgb & 0x00FC00) >> 5;
			pixel+=(rgb & 0x0000F8) >> 3;
			break;
		case 8: /* RGB = 888 */
			pixel = rgb & 0xFFFFFF;
			break;
	}
	return pixel;
}

GdkGC *get_gc_by_color(GdkWindow *win, guint32 color)
{
	GdkGCValues gc_val;
	GdkGCValuesMask gc_mask;

	gc_val.foreground.pixel = rgb_to_pixel(color);
	gc_val.line_style = GDK_LINE_ON_OFF_DASH;
	gc_mask = GDK_GC_FOREGROUND|GDK_GC_LINE_STYLE;
	return gdk_gc_new_with_values(win,&gc_val,gc_mask);
}

void update_dest()
{
	/* Finally let's do that polar mapping !!! */
	int i,j,xmax,ymax,modt,modt_max,modr,modr_max,modt_ok;
	unsigned int uintx,uinty;
	double cost,sint,angle,angle_step,x,y,dx,dy;
	GdkGC *gc=NULL;
	GdkPixmap *pix;
	
	if ((!im_src) || (!im_dest)) return;
	if (!gc) gc = get_gc_by_color((GdkWindow*)pix_src,0x00FFFF);
	gtk_pixmap_get(GTK_PIXMAP(widget_src),&pix,NULL);
	
	xmax=im_src->width;
	ymax=im_src->height;
	modt_max=dt/50;
	modr_max=dr/10;
	modt=0;
	angle = -M_PI;
	angle_step = 2*M_PI/dt;
	for (i=0; i<dt; i++)
	{
		cost = cos(angle);
		sint = sin(angle);
		x = (double)xi + sint*(double)ri;
		y = (double)yi - cost*(double)ri;
		dx = ((double)xo + sint*(double)ro - x)/dr;
		dy = ((double)yo - cost*(double)ro - y)/dr;
		modt_ok=(modt==0);
		if (modt_ok) modt=modt_max; else modt--;
		modr=modr_max;
		for (j=0; j<dr; j++, x+=dx, y+=dy)
		{
			uintx=(unsigned int)x;
			uinty=(unsigned int)y;
			if ((uintx<xmax) && (uinty<ymax))
			{
				if (modt_ok)
				{
					if (modr==0)
					{
						gdk_draw_point(pix,gc,uintx,uinty);
						modr=modr_max;
					}
					else
						modr--;
				}
				gdk_image_put_pixel(im_dest,i,j,gdk_image_get_pixel(im_src,uintx,uinty));
			}
		}
		angle+=angle_step;
	}
	gtk_widget_draw(widget_src,NULL);
	gtk_widget_draw(widget_dest,NULL);
}

void update_src()
{
	static GdkPixmap *pix=NULL;
	static GdkGC *gc1=NULL, *gc2=NULL;
	static gint w=0, h=0;
	gint src_w,src_h;
	
	if (!pix_src) return;
	gdk_window_get_size((GdkWindow*)pix_src,&src_w,&src_h);
	if ((!pix) || (w != src_w) || (h != src_h))
	{
		if (pix) gdk_pixmap_unref(pix);
		w = src_w;
		h = src_h;
		pix = gdk_pixmap_new(NULL,w,h,gdk_visual_get_best()->depth);
	}
	if (!gc1) gc1 = get_gc_by_color((GdkWindow*)pix_src,0xFF);
	if (!gc2) gc2 = get_gc_by_color((GdkWindow*)pix_src,0xFF0000);
	
	gdk_draw_pixmap(pix,gc1,pix_src,0,0,0,0,w,h);
	/* outer */
	gdk_draw_arc(pix,gc1,FALSE,xo-ro,yo-ro,ro*2,ro*2,0,360<<6);
	gdk_draw_line(pix,gc1,xo-CURSW,yo,xo+CURSW,yo);
	gdk_draw_line(pix,gc1,xo,yo-CURSW,xo,yo+CURSW);
	/* inner */
	gdk_draw_arc(pix,gc2,FALSE,xi-ri,yi-ri,ri*2,ri*2,0,360<<6);
	gdk_draw_line(pix,gc2,xi-CURSW,yi,xi+CURSW,yi);
	gdk_draw_line(pix,gc2,xi,yi-CURSW,xi,yi+CURSW);
	gtk_pixmap_set(GTK_PIXMAP(widget_src),pix,NULL);
	gtk_widget_draw(widget_src,NULL);
	update_dest();
}


void resize_dest(GtkSpinButton *spin, int *value)
{
	*value = gtk_spin_button_get_value_as_int(spin);
	if (im_dest) gdk_image_destroy(im_dest);
	im_dest = gdk_image_new(GDK_IMAGE_FASTEST,gdk_visual_get_best(),dt,dr);
	gtk_image_set (GTK_IMAGE(widget_dest),im_dest,NULL);
	update_dest();
}

void load_pic_by_name (char *fname)
{
	GdkImlibImage *im=NULL;
	gint w,h;
	int i,j,p;
	unsigned char *rgb;
		
	if (fname && (*fname != 0))
	{
		/* First get GdkImage */
		im=gdk_imlib_load_image(fname);
		im_src=gdk_image_new(GDK_IMAGE_FASTEST,gdk_visual_get_best(),im->rgb_width,im->rgb_height);
		rgb = im->rgb_data;
		p=0;
		for (j=0; j<im->rgb_height; j++)
			for (i=0; i<im->rgb_width; i++, p+=3)
				gdk_image_put_pixel(im_src,i,j,rgb_to_pixel((rgb[p]<<16)+(rgb[p+1]<<8)+rgb[p+2]));
		gdk_imlib_destroy_image(im);
		/* Then load into a GdkPixmap */
		gdk_imlib_load_file_to_pixmap(fname,&pix_src,NULL);
		gdk_window_get_size((GdkWindow*)pix_src,&w,&h);
		xo = w/2;
		yo = h/2;
		ro = (w+h)/4;
		xi = w/2 - CURSW;
		yi = h/2 - CURSW;
		ri = (w+h)/12;
		update_src();
	}
}

void load_pic_ok(GtkButton *button, GtkFileSelection *filesel)
{
	load_pic_by_name(gtk_file_selection_get_filename(filesel));
}

void load_pic(GtkButton *button, gpointer data)
{
	GtkWidget *filesel;
	
	filesel = gtk_file_selection_new("Load an image");
	gtk_signal_connect (GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
		"clicked",
		GTK_SIGNAL_FUNC(load_pic_ok),
		filesel);
	gtk_signal_connect_object (GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
		"clicked",
		GTK_SIGNAL_FUNC(gtk_widget_destroy),
		(gpointer)filesel);
	gtk_signal_connect_object (GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
		"clicked",
		GTK_SIGNAL_FUNC(gtk_widget_destroy),
		(gpointer)filesel);
	gtk_widget_show(filesel);
	gtk_window_set_modal(GTK_WINDOW(filesel), TRUE);
}


void save_pic_by_name (char *fname)
{
	FILE *f;
	char buf[1024];
	int i;
	
	sprintf(buf,"_%dx%d.raw",im_dest->width,im_dest->height);
	strcat(fname,buf);
	f=fopen(fname,"wb");
	for (i=0; i<im_dest->height; i++)
		fwrite((char*)(im_dest->mem)+im_dest->bpl*i, im_dest->width*im_dest->bpp, 1, f);
	fclose(f);
}

void save_pic_ok(GtkButton *button, GtkFileSelection *filesel)
{
	save_pic_by_name(gtk_file_selection_get_filename(filesel));
}

void save_pic(GtkButton *button, gpointer data)
{
	GtkWidget *filesel;
	
	filesel = gtk_file_selection_new("Save mapped image");
	gtk_signal_connect (GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
		"clicked",
		GTK_SIGNAL_FUNC(save_pic_ok),
		filesel);
	gtk_signal_connect_object (GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
		"clicked",
		GTK_SIGNAL_FUNC(gtk_widget_destroy),
		(gpointer)filesel);
	gtk_signal_connect_object (GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
		"clicked",
		GTK_SIGNAL_FUNC(gtk_widget_destroy),
		(gpointer)filesel);
	gtk_widget_show(filesel);
	gtk_window_set_modal(GTK_WINDOW(filesel), TRUE);
}


/* Interface */


gboolean src_clicked(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	int x,y;
	
	x=(int)event->x;
	y=(int)event->y;
	if ((abs(xo-x)<=CURSW) && (abs(yo-y)<=(CURSW/2))) src_grab=GRAB_XYO; else
	if ((abs(xi-x)<=CURSW) && (abs(yi-y)<=(CURSW/2))) src_grab=GRAB_XYI; else
	if (fabs(hypot(xo-x,yo-y) - ro) < 1.5) src_grab=GRAB_RO; else
	if (fabs(hypot(xi-x,yi-y) - ri) < 1.5) src_grab=GRAB_RI; else
		src_grab=GRAB_NONE;
/*	if (event->button == 3)
		switch (src_grab)
		{
			case GRAB_XYO: src_grab=GRAB_RO; break;
			case GRAB_XYI: src_grab=GRAB_RI; break;
			case GRAB_RO: src_grab=GRAB_XYO; break;
			case GRAB_RI: src_grab=GRAB_XYI; break;
		}
*/
	return TRUE;
}

gboolean src_unclicked(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	src_grab = GRAB_NONE;
	return TRUE;
}

gboolean src_moved(GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
	int x,y;
	
	x=(int)event->x;
	y=(int)event->y;
	switch (src_grab)
	{
		case GRAB_XYO: xo=x; yo=y; update_src(); break;
		case GRAB_XYI: xi=x; yi=y; update_src(); break;
		case GRAB_RO: ro=(int)hypot(xo-x,yo-y); update_src(); break;
		case GRAB_RI: ri=(int)hypot(xi-x,yi-y); update_src(); break;
	}
	return TRUE;
}


gboolean quitting(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  gtk_main_quit();
	return TRUE;
}


void setup_interface()
{
	GtkWidget *win, *hbox, *vbox, *widget;
	GtkAdjustment *adj;
	GdkPixmap *pix;
	
	win=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_realize(win);
	
	vbox=gtk_vbox_new(FALSE,0);
	gtk_signal_connect(GTK_OBJECT(win),"delete-event",GTK_SIGNAL_FUNC(quitting),NULL);
	gtk_container_add(GTK_CONTAINER(win),vbox);
	
	/* first horizontal pane : src & dest pics */
	widget=gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(vbox),widget);
	hbox=gtk_hbox_new(FALSE,8);
	gtk_container_set_border_width(GTK_CONTAINER(hbox),4);
	gtk_container_add(GTK_CONTAINER(widget),hbox);
	
	widget=gtk_event_box_new();
	gtk_container_add(GTK_CONTAINER(hbox),widget);
	gtk_widget_add_events(widget,GDK_BUTTON_PRESS|GDK_BUTTON_RELEASE|GDK_MOTION_NOTIFY);
	gtk_signal_connect(GTK_OBJECT(widget),"button-press-event",GTK_SIGNAL_FUNC(src_clicked),NULL);
	gtk_signal_connect(GTK_OBJECT(widget),"button-release-event",GTK_SIGNAL_FUNC(src_unclicked),NULL);
	gtk_signal_connect(GTK_OBJECT(widget),"motion-notify-event",GTK_SIGNAL_FUNC(src_moved),NULL);
	
	pix = gdk_pixmap_new(NULL,64,64,gdk_visual_get_best()->depth);
	gdk_draw_rectangle(pix,win->style->black_gc,TRUE,0,0,-1,-1);
	widget_src=gtk_pixmap_new(pix,NULL);
	gtk_container_add(GTK_CONTAINER(widget),widget_src);
	gdk_pixmap_unref(pix);
	
	im_dest = gdk_image_new(GDK_IMAGE_FASTEST,gdk_visual_get_best(),dt,dr);
	widget_dest=gtk_image_new(im_dest,NULL);
	gtk_container_add(GTK_CONTAINER(hbox),widget_dest);
	
	/* second horizontal pane : control widgets */
	hbox=gtk_hbox_new(FALSE,0);
	gtk_box_pack_end(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
	
	widget=gtk_button_new_with_label(" Load ");
	gtk_signal_connect(GTK_OBJECT(widget),"clicked",GTK_SIGNAL_FUNC(load_pic),NULL);
	gtk_container_add(GTK_CONTAINER(hbox),widget);
	
	widget=gtk_button_new_with_label(" Save ");
	gtk_signal_connect(GTK_OBJECT(widget),"clicked",GTK_SIGNAL_FUNC(save_pic),NULL);
	gtk_container_add(GTK_CONTAINER(hbox),widget);
	
	adj = (GtkAdjustment*) gtk_adjustment_new(dt,10,1000,1,10,10);
	widget=gtk_spin_button_new(adj, 1.1, 0);
	gtk_container_add(GTK_CONTAINER(hbox),widget);
	gtk_signal_connect(GTK_OBJECT(widget),"changed",GTK_SIGNAL_FUNC(resize_dest),(gpointer)&dt);
	
	adj = (GtkAdjustment*) gtk_adjustment_new(dr,10,1000,1,10,10);
	widget=gtk_spin_button_new(adj, 1.1, 0);
	gtk_container_add(GTK_CONTAINER(hbox),widget);
	gtk_signal_connect(GTK_OBJECT(widget),"changed",GTK_SIGNAL_FUNC(resize_dest),(gpointer)&dr);
	
	/* That's all folks */
	gtk_widget_show_all(win);
}


int main(int argc, char **argv)
{
	gtk_init(&argc, &argv);
	gdk_imlib_init();
	setup_interface();
	if (argc>=2) load_pic_by_name(argv[1]);
	gtk_main();
	return 0;
}
